from __future__ import print_function

import tempfile
import os

from mock import Mock, patch

# These are the tests from the pagure/ git repo.
# Run with:
# PYTHONPATH=.:/path/to/pagure/checkout nosetests dist_git_auth_tests.py
import pagure
import pagure.lib
import tests

import dist_git_auth


def setUp():
    tests.setUp()


def tearDown():
    tests.tearDown()


def patch_pdc(values):
    """ Decorator to patch the PDC calls to return values for this test.

    Args:
        values (dict): A dictionary where the keys are project fullnames
            (i.e. namespace/name), and the values are dicts with key branch
            names, and values their supported status.
            Note that as namespace, the PDC "type" is used, which lacks the "s"
            for rpms and modules.
            e.g.: {'rpm/test': {'f28': True, 'f27': False}}
    """
    def pdc_get_paged(_, global_component, type, name, fields):
        """ Function that emulates the pdc.get_paged call

        Args as provided by dist_git_auth's calls to it.

        Args:
            _ (anything): PDCClient internal
            global_component (string): package name
            type (string): The PDC "type": "rpm", "module", ...
            name (string): The branch name
            fields (list): Always ["active"]
        """
        fullname = '%s/%s' % (type, global_component)
        if fullname not in values:
            return []
        val = values[fullname]
        if name not in val:
            return []
        val = val[name]
        if val in (True, False):
            return [{"active": val}]
        # This case is used to emulate "weird" results
        return val

    def decorator(func):
        def test_wrapper(*args, **kwargs):
            with patch.object(dist_git_auth.PDCClient, '__getitem__'):
                with patch.object(dist_git_auth.PDCClient, 'get_paged',
                                  side_effect=pdc_get_paged):
                    return func(*args, **kwargs)
        return test_wrapper
    return decorator


class DistGitAuthTests(tests.Modeltests):
    """ Test DistGitAuth ACLs with Fedora config. """

    maxDiff = None

    def setUp(self):
        """ Set up the environment in which to run the tests. """
        super(DistGitAuthTests, self).setUp()

        pagure.config.config['ACL_DEBUG'] = True
        pagure.config.config.update(self.dga_config)

        self.dga = dist_git_auth.DistGitAuth()
        self.dga.info = Mock(side_effect=print)
        # We default to saying it's not a forced push
        dist_git_auth.is_forced_push = Mock(return_value=False)

        # Create an RCM user/group
        rcmuser = pagure.lib.model.User(
            user='releng',
            fullname='Release Engineering',
            token='aaabbbcd',
            default_email='rcm@local.local',
        )
        self.session.add(rcmuser)
        self.session.flush()
        rcmgroup = pagure.lib.model.PagureGroup(
            group_name='relenggroup',
            group_type='user',
            display_name='Releng group',
            user_id=rcmuser.id,
        )
        self.session.add(rcmgroup)
        self.session.flush()
        rcmuser.group_objs.append(rcmgroup)
        self.session.commit()

    def tearDown(self):
        """ Tear down the environment in which the tests ran. """
        self.dga = None
        super(DistGitAuthTests, self).tearDown()

    def create_namespaced_project(self, namespace, name, is_fork=False):
        item = pagure.lib.model.Project(
            user_id=1,  # pingou
            name=name,
            is_fork=is_fork,
            parent_id=3 if is_fork else None,
            description='namespaced test project',
            hook_token='aaabbbeee',
            namespace=namespace,
        )
        item.close_status = [
            'Invalid', 'Insufficient data', 'Fixed', 'Duplicate']
        self.session.add(item)
        self.session.commit()
        return pagure.lib._get_project(
            self.session, name=name, namespace=namespace)

    def expect_info_msg(self, expect_msg):
        found = False
        for call in self.dga.info.call_args_list:
            args = call[0]
            msg = args[0]
            if msg == expect_msg:
                found = True
        if not found:
            raise AssertionError(
                "Info message '%s' expected but not found" % expect_msg)


class DistGitAuthTestsGeneric(DistGitAuthTests):
    dga_config = {}

    def test_unused_repotype(self):
        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=None,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype='tickets',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Repotype tickets not in use")

    def test_branch_deletion(self):
        dist_git_auth.is_forced_push.return_value = True

        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto='0000000000000000000000000000000000000000',
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch deletion is not allowed")

    def test_forced_push(self):
        dist_git_auth.is_forced_push.return_value = True

        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Forced pushes are not allowed")

    def test_internal(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=True,
            )
        )

        self.expect_info_msg("Internal push allowed")

    def test_deploykey(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="deploykey_foobar",
                refname=None,
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Deploy keys are disabled")

    def test_invalid_user(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertRaises(
            pagure.exceptions.PagureException,
            self.dga.check_acl,
            self.session,
            project=project,
            username="nosuchuser",
            refname=None,
            pull_request=None,
            repodir=None,
            repotype='main',
            revfrom=None,
            revto=None,
            is_internal=False,
        )

    def test_unprotected_committer(self):
        project = self.create_namespaced_project('unprotected', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push to unprotected")

    def test_unprotected_non_committer(self):
        project = self.create_namespaced_project('unprotected', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Fall-through deny")

    def test_unprotected_pr_required_pr(self):
        project = self.create_namespaced_project('unprotected', 'test')
        self.dga.global_pr_only = True

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=True,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push to unprotected")

    def test_unprotected_pr_required_no_pr(self):
        project = self.create_namespaced_project('unprotected', 'test')
        self.dga.global_pr_only = True

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("A pull request is required for this branch")

    def test_unprotected_pr_required_requests(self):
        project = self.create_namespaced_project('unprotected', 'test')
        self.dga.global_pr_only = True

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='requests',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push to unprotected")


class DistGitAuthTestsFedora(DistGitAuthTests):
    dga_config = {
        'PR_ONLY': False,
        'ACL_BLOCK_UNSPECIFIED': False,
        'BLACKLIST_RES': ['refs/heads/c[0-9]+.*'],
        'UNSPECIFIED_BLACKLIST_RES': [
            'refs/heads/f[0-9]+',
        ],
        'RCM_GROUP': 'relenggroup',
        'RCM_BRANCHES': ['refs/heads/f[0-9]+'],
        'ACL_PROTECTED_NAMESPACES': ['rpms', 'modules', 'container'],
        'PDC_URL': 'invalid://',
    }

    def test_protected_blacklisted_ref(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Ref refs/heads/c7 is blocked")

    def test_protected_rcm(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="releng",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("RCM push")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_unsupported_branch(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f26",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f26 is unsupported")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_supported_branch_committer(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f27 is supported")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_supported_branch_non_committer(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f27 is supported")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_unspecified_branch_blacklisted(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f28",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified ref refs/heads/f28 is blocked")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_unspecified_branch_normal_committer(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified branch push")

    @patch_pdc({"rpm/test": {"f26": False, "f27": True}})
    def test_protected_unspecified_branch_normal_non_committer(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified branch push")


class DistGitAuthTestsCentOS(DistGitAuthTests):
    dga_config = {
        'PR_ONLY': False,
        'ACL_BLOCK_UNSPECIFIED': True,
        'BLACKLIST_RES': ['refs/heads/f[0-9]+.*'],
        'RCM_GROUP': 'relenggroup',
        'RCM_BRANCHES': ['refs/heads/c[0-9]+'],
        'SUPPORTED_SIGS': ['sig-core'],
        'SIG_PREFIXES': ['refs/heads/c7'],
        'ACL_PROTECTED_NAMESPACES': ['rpms'],
    }

    def setUp(self):
        super(DistGitAuthTestsCentOS, self).setUp()

        # Create an RCM user/group
        arrfab = pagure.lib.model.User(
            user='arrfab',
            fullname='Fabian Arriton',
            token='aaabbbcd',
            default_email='arrfab@local.local',
        )
        self.session.add(arrfab)
        self.session.flush()
        sigcore = pagure.lib.model.PagureGroup(
            group_name='sig-core',
            group_type='user',
            display_name='Core SIG group',
            user_id=arrfab.id,
        )
        self.session.add(sigcore)
        self.session.flush()
        arrfab.group_objs.append(sigcore)
        self.session.commit()

    def test_protected_blacklisted_ref(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Ref refs/heads/f27 is blocked")

    def test_protected_rcm(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="releng",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("RCM push")

    def test_protected_sig_sig_member(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="arrfab",
                refname="refs/heads/c7-sig-core-test",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("SIG push")

    def test_protected_sig_no_sig_member(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/c7-sig-core-test",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Access to namespace rpms is restricted")

    def test_protected_sig_sig_member_no_sig_branch(self):
        project = self.create_namespaced_project('rpms', 'test')

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="arrfab",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype='main',
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Access to namespace rpms is restricted")
